/*
 * eqBox.cpp
 *
 * Simple Equalizer example application for VR environments.
 * Based on the Equalizer eqHello example.
 *
 * Copyright (c) 2007, Stefan Eilemann <eile@equalizergraphics.com>
 * Copyright (C) 2008, 2009, 2010, 2011, 2012
 * Computer Graphics Group, University of Siegen, Germany.
 * See http://www.cg.informatik.uni-siegen.de/ for contact information.
 * Written by Martin Lambers <martin.lambers@uni-siegen.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of Eyescale Software GmbH nor the names of its
 *   contributors may be used to endorse or promote products derived from this
 *   software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */


#include <iostream>
#include <cstring>

#include <eq/eq.h>


/*
 * EqInitData
 *
 * Data that is set only once and remains constant for all frames.
 */

class EqInitData : public co::Object
{
public:
    struct {
        eq::uint128_t frame_data_id;
        float light_pos[4];
        float light_ambient[4];
        float rotation_axis[3];
    } data;

    EqInitData();
    void set_frame_data_id(const eq::uint128_t &id);
    eq::uint128_t get_frame_data_id() const;

protected:
    virtual void getInstanceData(co::DataOStream &os);
    virtual void applyInstanceData(co::DataIStream &is);
    virtual ChangeType getChangeType() const;
};

EqInitData::EqInitData() : co::Object()
{
}

void EqInitData::set_frame_data_id(const eq::uint128_t &id)
{
    data.frame_data_id = id;
}

eq::uint128_t EqInitData::get_frame_data_id() const
{
    return data.frame_data_id;
}

void EqInitData::getInstanceData(co::DataOStream &os)
{
    os.write(&data, sizeof(data));
}

void EqInitData::applyInstanceData(co::DataIStream &is)
{
    is.read(&data, sizeof(data));
}

co::Object::ChangeType EqInitData::getChangeType() const
{
    return STATIC;
}


/*
 * EqFrameData
 *
 * Data that may change with each new frame.
 */

class EqFrameData : public co::Object
{
public:
    struct {
        float rotation_angle;
        float head_pos[3];
        float boxPos[3];
    } data;

    EqFrameData();

protected:
    virtual void getInstanceData(co::DataOStream &os);
    virtual void applyInstanceData(co::DataIStream &is);
    virtual ChangeType getChangeType() const;
};

EqFrameData::EqFrameData() : co::Object()
{
    data.rotation_angle = 0.0f;
    data.head_pos[0] = 0.0f;
    data.head_pos[1] = 0.0f;
    data.head_pos[2] = 0.0f;
    data.boxPos[0] = 0.0f;
    data.boxPos[1] = 0.0f;
    data.boxPos[2] = -5.0f;
}

void EqFrameData::getInstanceData(co::DataOStream &os)
{
    os.write(&data, sizeof(data));
}

void EqFrameData::applyInstanceData(co::DataIStream &is)
{
    is.read(&data, sizeof(data));
}

co::Object::ChangeType EqFrameData::getChangeType() const
{
    return INSTANCE;
}


/*
 * EqConfig
 *
 * Controls all instances and handles events.
 */

class EqConfig : public eq::Config
{
private:
    EqInitData init_data;               // Master EqInitData instance
    EqFrameData frame_data;             // Master EqFrameData instance
    int64_t last_animation;             // Time stamp of last object animation (milliseconds)

public:
    EqConfig(eq::ServerPtr parent);
    bool init(const EqInitData &init_data_template);
    virtual bool exit();
    virtual eq::uint128_t startFrame();
    virtual bool handleEvent(const eq::ConfigEvent *event);
};

EqConfig::EqConfig(eq::ServerPtr parent) : eq::Config(parent)
{
}

bool EqConfig::init(const EqInitData &init_data_template)
{
    last_animation = getTime();
    // Initialize EqInitData master instance
    init_data.data = init_data_template.data;
    // Register master instances
    registerObject(&frame_data);
    init_data.set_frame_data_id(frame_data.getID());
    registerObject(&init_data);
    return eq::Config::init(init_data.getID());
}

bool EqConfig::exit()
{
    const bool ret = eq::Config::exit();
    // Deregister master instances
    deregisterObject(&init_data);
    deregisterObject(&frame_data);
    return ret;
}

eq::uint128_t EqConfig::startFrame()
{
    // Animate: Update rotation
    int64_t now = getTime();
    if (now - last_animation > 10) {
        frame_data.data.rotation_angle += 0.1f;
        last_animation = now;
    }

    // Set viewer head position and orientation
    // TODO: In a VR environment, these values should be read from a tracking device.
    eq::Matrix4f head_matrix = eq::Matrix4f::IDENTITY;
    head_matrix.set_translation(frame_data.data.head_pos[0], frame_data.data.head_pos[1], frame_data.data.head_pos[2]);
    getObservers().at(0)->setHeadMatrix(head_matrix);

    // Commit new version of updated frame data
    const eq::uint128_t version = frame_data.commit();

    // Start this frame with the committed frame data
    bool ret = eq::Config::startFrame(version);

    return ret;
}

bool EqConfig::handleEvent(const eq::ConfigEvent *event)
{
    if (eq::Config::handleEvent(event))
        return true;

    // Handle keyboard and/or mouse events here.
    // Such events usually update the frame data.
    if (event->data.type == eq::Event::KEY_PRESS) {
        switch (event->data.keyPress.key) {
        case eq::KC_UP:
            frame_data.data.boxPos[1] += 0.1f;
            break;

        case eq::KC_DOWN:
            frame_data.data.boxPos[1] -= 0.1f;
            break;

        case eq::KC_LEFT:
            frame_data.data.boxPos[0] -= 0.1f;
            break;

        case eq::KC_RIGHT:
            frame_data.data.boxPos[0] += 0.1f;
            break;
        }
    }

    return true;
}


/*
 * EqNode
 *
 * Each node has a slave instance of EqInitData.
 */

class EqNode : public eq::Node
{
private:
    EqInitData init_data;

public:
    EqNode(eq::Config *parent);
    const EqInitData &get_init_data() const;

protected:
    virtual bool configInit(const eq::uint128_t &init_id);
    virtual bool configExit();
    virtual void frameStart(const eq::uint128_t &frame_id, const uint32_t frame_number);
    virtual void frameFinish(const eq::uint128_t &frame_id, const uint32_t frame_number);
};

EqNode::EqNode(eq::Config *parent) : eq::Node(parent)
{
}

const EqInitData &EqNode::get_init_data() const
{
    return init_data;
}

bool EqNode::configInit(const eq::uint128_t &init_id)
{
    if (!eq::Node::configInit(init_id))
        return false;
    // Map our EqInitData instance to the master instance
    eq::Config *config = getConfig();
    return config->mapObject(&init_data, init_id);
}

bool EqNode::configExit()
{
    // Unmap our EqInitData instance
    eq::Config *config = getConfig();
    config->unmapObject(&init_data);
    return eq::Node::configExit();
}

void EqNode::frameStart(const eq::uint128_t &/* frame_id */, const uint32_t frame_number)
{
    // Equalizer calls this function at the beginning of each frame.
    startFrame(frame_number);
}

void EqNode::frameFinish(const eq::uint128_t &/* frame_id */, const uint32_t frame_number)
{
    // Equalizer calls this function at the end of each frame.
    releaseFrame(frame_number);
}


/*
 * EqPipe
 *
 * Each pipe has a slave instance of EqFrameData.
 */

class EqPipe : public eq::Pipe
{
private:
    EqFrameData frame_data;

public:
    EqPipe(eq::Node *parent);
    const EqFrameData &get_frame_data() const;

protected:
    virtual bool configInit(const eq::uint128_t &init_id);
    virtual bool configExit();
    virtual void frameStart(const eq::uint128_t &frame_id, const uint32_t frame_number);
    virtual void frameFinish(const eq::uint128_t &frame_id, const uint32_t frame_number);
};

EqPipe::EqPipe(eq::Node *parent) : eq::Pipe(parent)
{
}

const EqFrameData &EqPipe::get_frame_data() const
{
    return frame_data;
}

bool EqPipe::configInit(const eq::uint128_t &init_id)
{
    if (!eq::Pipe::configInit(init_id))
        return false;
    // Map our EqFrameData instance to the master instance
    const EqNode *node = static_cast<EqNode *>(getNode());
    const EqInitData &init_data = node->get_init_data();
    const eq::uint128_t frame_data_id = init_data.get_frame_data_id();
    eq::Config *config = getConfig();
    return config->mapObject(&frame_data, frame_data_id);
}

bool EqPipe::configExit()
{
    // Unmap our EqFrameData instance
    eq::Config *config = getConfig();
    config->unmapObject(&frame_data);
    return eq::Pipe::configExit();
}

void EqPipe::frameStart(const eq::uint128_t &frame_id, const uint32_t frame_number)
{
    // Equalizer calls this function at the beginning of each frame.
    // Update our EqFrameData instance from the master instance before drawing
    frame_data.sync(frame_id);
    startFrame(frame_number);
}

void EqPipe::frameFinish(const eq::uint128_t &/* frame_id */, const uint32_t frame_number)
{
    // Equalizer calls this function at the end of each frame.
    releaseFrame(frame_number);
}


/*
 * EqWindow
 *
 * The window has an OpenGL context, and manages context-specific data.
 */

class EqWindow : public eq::Window
{
public:
    EqWindow(eq::Pipe *parent);
    GLuint box_display_list;

protected:
    virtual bool configInitGL(const eq::uint128_t &init_id);
    virtual bool configExitGL();
    virtual void frameStart(const eq::uint128_t &frame_id, const uint32_t frame_number);
    virtual void frameFinish(const eq::uint128_t &frame_id, const uint32_t frame_number);
};

EqWindow::EqWindow(eq::Pipe *parent) : eq::Window(parent)
{
}

bool EqWindow::configInitGL(const eq::uint128_t &init_id)
{
    if (!eq::Window::configInitGL(init_id))
        return false;

    // Initialize data objects that are OpenGL context specific here,
    // e.g. textures and display lists.

    box_display_list = glGenLists(1);
    glNewList(box_display_list, GL_COMPILE);
    for (int i = 0; i < 6; i++) {
        glColor3f(
                (i & 1) ? 0.5f : 1.0f,
                (i & 2) ? 1.0f : 0.5f,
                (i & 4) ? 1.0f : 0.5f);
        glNormal3f(0.0f, 0.0f, 1.0f);
        glBegin(GL_TRIANGLE_STRIP);
        glVertex3f( 0.7f,  0.7f, -1.0f);
        glVertex3f(-0.7f,  0.7f, -1.0f);
        glVertex3f( 0.7f, -0.7f, -1.0f);
        glVertex3f(-0.7f, -0.7f, -1.0f);
        glEnd();
        if (i < 3)
            glRotatef(90.0f, 0.0f, 1.0f, 0.0f);
        else if (i == 3)
            glRotatef(90.0f, 1.0f, 0.0f, 0.0f);
        else
            glRotatef(180.0f, 1.0f, 0.0f, 0.0f);
    }
    glEndList();

    return true;
}

bool EqWindow::configExitGL()
{
    return eq::Window::configExitGL();
}

void EqWindow::frameStart(const eq::uint128_t &/* frame_id */, const uint32_t frame_number)
{
    // Equalizer calls this function at the beginning of each frame.
    startFrame(frame_number);
}

void EqWindow::frameFinish(const eq::uint128_t &/* frame_id */, const uint32_t frame_number)
{
    // Equalizer calls this function at the end of each frame.
    releaseFrame(frame_number);
}


/*
 * EqChannel
 */

class EqChannel : public eq::Channel
{
public:
    EqChannel(eq::Window *parent);

protected:
    virtual bool configInit(const eq::uint128_t &init_id);
    virtual void frameDraw(const eq::uint128_t &frame_id);
    virtual void frameStart(const eq::uint128_t &frame_id, const uint32_t frame_number);
    virtual void frameFinish(const eq::uint128_t &frame_id, const uint32_t frame_number);
};

EqChannel::EqChannel(eq::Window *parent) : eq::Channel(parent)
{
}

bool EqChannel::configInit(const eq::uint128_t &init_id)
{
    if (!eq::Channel::configInit(init_id))
        return false;
    // No configuration tasks required in this example.
    return true;
}

void EqChannel::frameDraw(const eq::uint128_t &frame_id)
{
    // Let Equalizer set the projection matrix (and other things)
    eq::Channel::frameDraw(frame_id);

    // Get init data and frame data
    const EqNode *node = static_cast<EqNode *>(getNode());
    const EqInitData &init_data = node->get_init_data();
    const EqPipe *pipe = static_cast<EqPipe *>(getPipe());
    const EqFrameData &frame_data = pipe->get_frame_data();
    const EqWindow *window = static_cast<EqWindow *>(getWindow());

    // Set lights
    glLightfv(GL_LIGHT0, GL_POSITION, init_data.data.light_pos);
    glLightfv(GL_LIGHT0, GL_AMBIENT, init_data.data.light_ambient);

    // Set translation
    glTranslatef(
            frame_data.data.boxPos[0],
            frame_data.data.boxPos[1],
            frame_data.data.boxPos[2]);

    // Set rotation
    glRotatef(
            frame_data.data.rotation_angle,
            init_data.data.rotation_axis[0],
            init_data.data.rotation_axis[1],
            init_data.data.rotation_axis[2]);

    // Render the box
    glCallList(window->box_display_list);
}

void EqChannel::frameStart(const eq::uint128_t &/* frame_id */, const uint32_t frame_number)
{
    // Equalizer calls this function at the beginning of each frame.
    startFrame(frame_number);
}

void EqChannel::frameFinish(const eq::uint128_t &/* frame_id */, const uint32_t frame_number)
{
    // Equalizer calls this function at the end of each frame.
    releaseFrame(frame_number);
}


/*
 * EqNodeFactory
 *
 * Used by Equalizer to create instances of our classes.
 */

class EqNodeFactory : public eq::NodeFactory
{
public:
    virtual eq::Config *createConfig(eq::ServerPtr parent) { return new EqConfig(parent); }
    virtual eq::Node *createNode(eq::Config *parent) { return new EqNode(parent); }
    virtual eq::Pipe *createPipe(eq::Node *parent) { return new EqPipe(parent); }
    virtual eq::Window *createWindow(eq::Pipe *parent) { return new EqWindow(parent); }
    virtual eq::Channel *createChannel(eq::Window *parent) { return new EqChannel(parent); }
};


/*
 * Main
 */

int main(int argc, char *argv[])
{
    /* Help text */
    std::cout << "The arrow keys (up, down, left, right) move the box." << std::endl;
    std::cout << "ESC quits the program." << std::endl;

    /* Initialize Equalizer */
    EqNodeFactory node_factory;
    if (!eq::init(argc, argv, &node_factory)) {
        std::cerr << "Equalizer initialization failed!" << std::endl;
        return 1;
    }

    /* Initialize EqInitData. This may be influenced e.g. by command line options. */
    EqInitData init_data;
    init_data.data.light_pos[0] = 0.0f;
    init_data.data.light_pos[1] = 0.0f;
    init_data.data.light_pos[2] = 1.0f;
    init_data.data.light_pos[3] = 0.0f;
    init_data.data.light_ambient[0] = 0.2f;
    init_data.data.light_ambient[1] = 0.2f;
    init_data.data.light_ambient[2] = 0.2f;
    init_data.data.light_ambient[3] = 1.0f;
    init_data.data.rotation_axis[0] = 1.0f;
    init_data.data.rotation_axis[1] = 0.5f;
    init_data.data.rotation_axis[2] = 0.25f;

    /* Get a configuration */
    bool error = false;
    EqConfig *config = static_cast<EqConfig *>(eq::getConfig(argc, argv));
    // The following code is only executed on the application node because
    // eq::getConfig() does not return on other nodes.
    if (config) {
        /* Initialize configuration */
        if (config->init(init_data)) {
            /* Run main loop */
            while (config->isRunning()) {
                config->startFrame();
                config->finishFrame();
            }
            /* Exit configuration */
            config->exit();
        } else {
            std::cerr << "EqConfig initialization failed" << std::endl;
            error = true;
        }
        /* Release configuration */
        eq::releaseConfig(config);
    } else {
        std::cerr << "Cannot get configuration" << std::endl;
        error = true;
    }

    /* Exit */
    eq::exit();
    return error ? 1 : 0;
}
